class MetadataProvider(object):

    def fetch_org_metadata(self,org_id):
        pass


    def fetch_service_metadata(self,org_id,service_id):
        pass

    def enhance_service_group_details(self, service_metadata, org_metadata):
        pass

